!-----------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations         !
!   Copyright (C) 2000 - 2009  CP2K developers group                          !
!-----------------------------------------------------------------------------!

! *****************************************************************************
!> \brief Rountines to calculate HFX energy and potential 
!> \par History
!>      11.2006 created [Manuel Guidon]
!> \author Manuel Guidon 
! *****************************************************************************
MODULE hfx_energy_potential 
  
  USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                             get_atomic_kind_set
  USE bibliography,                    ONLY: cite_reference,&
                                             guidon2008
  USE cell_types,                      ONLY: cell_type,&
                                             pbc
  USE cp_files,                        ONLY: get_unit_number
  USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                             cp_print_key_unit_nr
  USE cp_para_types,                   ONLY: cp_para_env_type
  USE gamma,                           ONLY: init_md_ftable
  USE hfx_communication,               ONLY: distribute_ks_matrix,&
                                             get_full_density,&
                                             get_is_assoc_atomic_block
  USE hfx_compression_methods,         ONLY: hfx_add_mult_cache_elements,&
                                             hfx_add_single_cache_element,&
                                             hfx_decompress_first_cache,&
                                             hfx_flush_last_cache,&
                                             hfx_get_mult_cache_elements,&
                                             hfx_get_single_cache_element,&
                                             hfx_reset_cache_and_container
  USE hfx_libint_interface,            ONLY: evaluate_eri,&
                                             set_eps_cutoff
  USE hfx_libint_wrapper,              ONLY: initialize_libint,&
                                             lib_int
  USE hfx_load_balance_methods,        ONLY: build_pair_list,&
                                             load_balance_energy,&
                                             update_load_balance_energy
  USE hfx_screening_methods,           ONLY: calc_pair_dist_radii,&
                                             calc_screening_functions,&
                                             update_pmax_mat
  USE hfx_types,                       ONLY: &
       hfx_basis_info_type, hfx_basis_type, hfx_cache_type, hfx_cell_type, &
       hfx_container_type, hfx_general_type, hfx_init_container, &
       hfx_load_balance_type, hfx_memory_type, hfx_potential_type, &
       hfx_reset_memory_usage_counter, hfx_screen_coeff_type, &
       hfx_screening_type, hfx_type, log_zero, pair_list_type, &
       pair_set_list_type
  USE input_constants,                 ONLY: do_hfx_potential_truncated
  USE input_section_types,             ONLY: section_vals_type
  USE kinds,                           ONLY: dp,&
                                             int_8
  USE machine,                         ONLY: m_memory,&
                                             m_walltime
  USE mathconstants
  USE message_passing,                 ONLY: mp_max_i8,&
                                             mp_sum,&
                                             mp_sum_i8
  USE orbital_pointers
  USE particle_types,                  ONLY: particle_type
  USE qs_energy_types,                 ONLY: qs_energy_type
  USE qs_environment_types,            ONLY: get_qs_env,&
                                             qs_environment_type
  USE qs_rho_types,                    ONLY: qs_rho_type
  USE sparse_matrix_types,             ONLY: cp_sm_sm_trace,&
                                             real_matrix_p_type
  USE t_c_g0,                          ONLY: init
  USE timings,                         ONLY: timeset,&
                                             timestop
#include "cp_common_uses.h"

  IMPLICIT NONE
  PRIVATE
   
  PUBLIC  integrate_four_center
  INTEGER, SAVE                                         :: init_t_c_g0_lmax = -1

  CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'hfx_energy_potential'

!***
  
  CONTAINS

! *****************************************************************************
!> \brief computes four center integrals for a full basis set and updates the
!>      Kohn-Sham-Matrix and energy. Uses all 8 eri symmetries
!> \param qs_env  
!> \param ks_matrix
!> \param energy
!> \param rho density matrix
!> \param hfx_section input_section HFX
!> \param para_env
!> \param geometry_did_change flag that indicates we have to recalc integrals
!> \param irep Index for the HFX replica
!> \param distribute_fock_matrix Flag that indicates whether to communicate the
!>        new fock matrix or not
!> \param error variable to control error logging, stopping,...
!>        see module cp_error_handling
!> \par History
!>      06.2007 created [Manuel Guidon]
!>      08.2007 optimized load balance [Manuel Guidon] 
!>      09.2007 new parallelization [Manuel Guidon]
!>      02.2009 completely rewritten screening part [Manuel Guidon]
!> \author Manuel Guidon
! *****************************************************************************
  SUBROUTINE integrate_four_center(qs_env,ks_matrix,energy,rho,hfx_section,para_env,& 
                                   geometry_did_change,irep,distribute_fock_matrix,& 
                                   ispin, error)

    TYPE(qs_environment_type), POINTER       :: qs_env
    TYPE(real_matrix_p_type), DIMENSION(:), &
      POINTER                                :: ks_matrix
    TYPE(qs_energy_type), POINTER            :: energy
    TYPE(qs_rho_type), POINTER               :: rho
    TYPE(section_vals_type), POINTER         :: hfx_section
    TYPE(cp_para_env_type), POINTER          :: para_env
    LOGICAL                                  :: geometry_did_change
    INTEGER                                  :: irep
    LOGICAL, INTENT(IN)                      :: distribute_fock_matrix
    INTEGER, INTENT(IN)                      :: ispin
    TYPE(cp_error_type), INTENT(inout)       :: error

    CHARACTER(LEN=*), PARAMETER :: routineN = 'integrate_four_center', &
      routineP = moduleN//':'//routineN

    INTEGER :: bin, buffer_left, buffer_size, buffer_start, cache_size, &
      current_counter, handle, handle_dist_ks, handle_getP, handle_load, &
      handle_main, i, i_thread, iatom, ikind, iset, iw, jatom, jkind, jset, &
      katom, kkind, kset, latom, lkind, lset, max_am, max_set, maxsgf, mb, &
      my_bin_size, n_threads, natom, nbits, nco_max, ncoa, ncob, ncoc, ncod, &
      nints, nseta, nsetb, nsetc, nsetd, nsgf_max, nsgfa_total, nsgfb_total, &
      nsgfc_total, nsgfd_total, nspins, pa1, pb1, pc1, pd1, sgfa, sgfb, sgfc, &
      sgfd, stat
    INTEGER(int_8) :: estimate_to_store_int, mem_compression_counter, &
      mem_compression_counter_disk, mem_eris, mem_eris_disk, mem_max_val, &
      mem_maxvals, my_current_counter, my_istart, my_original_cpu_id, &
      n_processes, ncpu, neris_disk, neris_incore, neris_onthefly, &
      neris_total, stor_count_int_disk, stor_count_max_val, &
      storage_counter_integrals
    INTEGER, ALLOCATABLE, DIMENSION(:)       :: kind_of, last_sgf_global
    INTEGER, DIMENSION(:), POINTER :: la_max, la_min, lb_max, lb_min, lc_max, &
      lc_min, ld_max, ld_min, npgfa, npgfb, npgfc, npgfd, nsgfa, nsgfb, &
      nsgfc, nsgfd
    INTEGER, DIMENSION(:, :), POINTER        :: first_sgfa, first_sgfb, &
                                                first_sgfc, first_sgfd
    LOGICAL                                  :: buffer_overflow, &
                                                do_p_screening, do_periodic, &
                                                failure
    REAL(dp) :: bintime_start, bintime_stop, cartesian_estimate, &
      compression_factor, compression_factor_disk, ene_x_aa, ene_x_bb, &
      eps_schwarz, eps_storage, fac, hf_fraction, max_contraction_val, &
      max_val1, max_val2, pmax_entry, ra(3), rb(3), rc(3), rd(3), &
      spherical_estimate, symm_fac
    REAL(dp), ALLOCATABLE, DIMENSION(:)      :: primitive_integrals, T1
    REAL(dp), DIMENSION(:, :), POINTER       :: max_contraction, sphi_a, &
                                                sphi_b, sphi_c, sphi_d, zeta, &
                                                zetb, zetc, zetd
    TYPE(atomic_kind_type), DIMENSION(:), &
      POINTER                                :: atomic_kind_set
    TYPE(cp_logger_type), POINTER            :: logger
    TYPE(hfx_basis_info_type), POINTER       :: basis_info
    TYPE(hfx_basis_type), DIMENSION(:), &
      POINTER                                :: basis_parameter
    TYPE(hfx_cache_type), DIMENSION(:), &
      POINTER                                :: integral_caches, &
                                                integral_caches_disk
    TYPE(hfx_cache_type), POINTER            :: maxval_cache, &
                                                maxval_cache_disk
    TYPE(hfx_container_type), DIMENSION(:), &
      POINTER                                :: integral_containers, &
                                                integral_containers_disk
    TYPE(hfx_container_type), POINTER        :: maxval_container, &
                                                maxval_container_disk
    TYPE(hfx_general_type)                   :: general_parameter
    TYPE(hfx_load_balance_type)              :: load_balance_parameter
    TYPE(hfx_memory_type), POINTER           :: memory_parameter
    TYPE(hfx_potential_type)                 :: potential_parameter
    TYPE(hfx_screening_type)                 :: screening_parameter
    TYPE(hfx_type), POINTER                  :: actual_x_data
    TYPE(particle_type), DIMENSION(:), &
      POINTER                                :: particle_set

!$  INTEGER :: omp_get_max_threads,omp_get_thread_num
    REAL(dp), DIMENSION(:, :, :, :), &
              POINTER, SAVE                  :: shm_initial_p
    REAL(dp), DIMENSION(:, :, :, :), &
              POINTER, SAVE                  :: shm_actual_pmax_set
    INTEGER, SAVE                            :: shm_number_of_p_entries
    INTEGER, DIMENSION(:,:), POINTER, SAVE   :: shm_is_assoc_atomic_block
    INTEGER(int_8)                           :: shm_neris_total, shm_neris_onthefly, &
                                                shm_storage_counter_integrals, shm_neris_incore, &
                                                shm_neris_disk, shm_stor_count_int_disk, shm_stor_count_max_val
    LOGICAL                                  :: do_disk_storage, use_disk_storage
    INTEGER(int_8)                           :: counter, neris_tmp, nprim_ints, shm_nprim_ints,&
                                                mb_size_p, mb_size_f, mb_size_s, mb_size_buffers
    INTEGER                                  :: nkind, l_max, istat
    REAL(dp)                                 :: ks_occ
    TYPE(cell_type), POINTER                 :: cell
    LOGICAL                                  :: exist
    CHARACTER(LEN=512)                       :: error_msg
    INTEGER                                  :: unit_id
    TYPE(hfx_screen_coeff_type), &
      DIMENSION(:,:,:,:,:,:), POINTER        :: screen_coeffs_pgf, radii_pgf
    TYPE(hfx_screen_coeff_type), &
      DIMENSION(:,:,:,:), POINTER            :: screen_coeffs_set
    TYPE(hfx_screen_coeff_type), &
      DIMENSION(:,:), POINTER                :: screen_coeffs_kind
    REAL(dp)                                 :: rab, rab2, rcd, rcd2, max_val1_from_funct,&
                                                max_val2_from_funct, log10_eps_schwarz, &
                                                B11(3), pbc_B(3), D11(3), pbc_D(3), log10_pmax, oops,&
                                                diag1, diag2, diag(3), min_diag(3), R_diag,ln_10
    INTEGER(int_8)                           :: subtr_size_mb, hash_idx

    TYPE(hfx_screen_coeff_type), &
      DIMENSION(:,:), POINTER                :: tmp_R_1, tmp_R_2, tmp_screen_pgf1, tmp_screen_pgf2

    INTEGER                                  :: max_pgf
    INTEGER(int_8)                           :: nanotime, time_1, time_2,memsize_before,memsize_after,&
                                                max_val_memory, max_val_memory_tmp, N_pmat,tmp_i8(8)
    REAL(dp)                                 :: BD2, AC2, min_PQ2, max_val2_set
    TYPE(lib_int)                            :: private_lib
    INTEGER                                  :: nrow, ncol, pa, pd
    REAL(dp), DIMENSION(:), POINTER          :: full_density, full_density_beta, full_ks,&
                                                full_ks_beta
    REAL(dp), DIMENSION(:), ALLOCATABLE      :: pbd_buf, pbc_buf, pad_buf, pac_buf,&
                                                kbd_buf, kbc_buf, kad_buf, kac_buf
    LOGICAL                                  :: treat_lsd_in_core, ks_fully_occ
    INTEGER                                  :: iatom_block, jatom_block, katom_block, latom_block, &
                                                block_size
    INTEGER(int_8)                           :: atom_block, tmp_block, nblocks
    INTEGER                                  :: i_list_ij,i_list_kl, &
                                                i_set_list_ij, i_set_list_ij_start, i_set_list_ij_stop,&
                                                i_set_list_kl, i_set_list_kl_start, i_set_list_kl_stop
    TYPE(pair_list_type)                     :: list_ij, list_kl
    REAL(KIND=dp)                            :: coeffs_kind_max0
    TYPE(pair_set_list_type), DIMENSION(:), ALLOCATABLE :: set_list_ij, set_list_kl


    CALL timeset(routineN,handle)

    CALL cite_reference(Guidon2008)


    ! ** This is not very clean, but effective. ispin can only be 2 if we do the beta spin part in core
    IF( ispin == 2 ) geometry_did_change = .FALSE.

    logger => cp_error_get_logger(error)

    IF( geometry_did_change ) THEN
      memsize_before=m_memory()
      CALL mp_max_i8(memsize_before, para_env%group)
      iw = cp_print_key_unit_nr(logger,hfx_section,"HF_INFO",&
                                extension=".scfLog",error=error)
      IF (iw>0) THEN
        WRITE (UNIT=iw,FMT="(/,(T3,A,T60,I20))")&
              "HFX_MEM_INFO| Est. max. program size before HFX [MB's]:", memsize_before/(1024*1024)
      END IF
      CALL cp_print_key_finished_output(iw,logger,hfx_section,&
                                       "HF_INFO", error=error)
    END IF

    CALL get_qs_env(qs_env=qs_env, atomic_kind_set=atomic_kind_set, cell=cell, error=error)

    !! Calculate l_max used in fgamma , because init_md_ftable is definitely not thread safe
    nkind = SIZE(atomic_kind_set,1)
    l_max = 0
    DO ikind=1,nkind
      l_max = MAX(l_max,MAXVAL(qs_env%x_data(1,1)%basis_parameter(ikind)%lmax))
    ENDDO
    l_max = 4*l_max 
    CALL init_md_ftable(l_max)

    IF(qs_env%x_data(1,1)%potential_parameter%potential_type == do_hfx_potential_truncated) THEN
      IF(l_max>init_t_c_g0_lmax) THEN
        IF( para_env%mepos == 0 ) THEN
          unit_id = get_unit_number()
          OPEN(unit_id,FILE=qs_env%x_data(1,1)%potential_parameter%filename)
        END IF
        CALL init(l_max, unit_id, para_env%mepos, para_env%group)
        IF(para_env%mepos == 0 ) THEN
          CLOSE(unit_id)
        END IF
        init_t_c_g0_lmax = l_max
      END IF
    CALL set_eps_cutoff(qs_env%x_data(1,1)%screening_parameter%eps_schwarz)
    END IF


    n_threads = 1
!$  n_threads = omp_get_max_threads()

    shm_neris_total = 0 
    shm_nprim_ints = 0
    shm_neris_onthefly = 0
    shm_storage_counter_integrals = 0
    shm_stor_count_int_disk = 0
    shm_neris_incore = 0
    shm_neris_disk = 0
    shm_stor_count_max_val = 0

!$OMP PARALLEL DEFAULT(PRIVATE) SHARED(qs_env,&
!$OMP                                  ks_matrix,&
!$OMP                                  energy,&
!$OMP                                  rho,&
!$OMP                                  hfx_section,&
!$OMP                                  para_env,&
!$OMP                                  geometry_did_change,&
!$OMP                                  irep,&
!$OMP                                  distribute_fock_matrix,&
!$OMP                                  error,&
!$OMP                                  ncoset,&
!$OMP                                  full_ks,&
!$OMP                                  full_ks_beta,&
!$OMP                                  n_threads,&
!$OMP                                  full_density,&
!$OMP                                  full_density_beta,&
!$OMP                                  shm_initial_p,&
!$OMP                                  shm_actual_pmax_set,&
!$OMP                                  shm_is_assoc_atomic_block,&
!$OMP                                  shm_number_of_p_entries,&
!$OMP                                  shm_neris_total,&
!$OMP                                  shm_neris_onthefly,&
!$OMP                                  shm_storage_counter_integrals,&
!$OMP                                  shm_stor_count_int_disk,&
!$OMP                                  shm_neris_incore,&
!$OMP                                  shm_neris_disk,&
!$OMP                                  shm_nprim_ints,&
!$OMP                                  shm_stor_count_max_val,&
!$OMP                                  cell,&
!$OMP                                  screen_coeffs_set,&
!$OMP                                  screen_coeffs_kind,&
!$OMP                                  screen_coeffs_pgf,&
!$OMP                                  radii_pgf,&
!$OMP                                  nkind,&
!$OMP                                  ispin)


    ln_10 = LOG(10.0_dp)
    i_thread = 0
!$  i_thread = omp_get_thread_num()
    
    actual_x_data => qs_env%x_data(irep, i_thread + 1)
    do_periodic = actual_x_data%periodic_parameter%do_periodic

    screening_parameter    = actual_x_data%screening_parameter
    potential_parameter    = actual_x_data%potential_parameter

    general_parameter      = actual_x_data%general_parameter
    load_balance_parameter = actual_x_data%load_balance_parameter
    memory_parameter       => actual_x_data%memory_parameter

    maxval_container       => actual_x_data%maxval_container
    maxval_cache           => actual_x_data%maxval_cache

    integral_containers    => actual_x_data%integral_containers
    integral_caches        => actual_x_data%integral_caches

    maxval_container_disk       => actual_x_data%maxval_container_disk
    maxval_cache_disk           => actual_x_data%maxval_cache_disk

    integral_containers_disk    => actual_x_data%integral_containers_disk
    integral_caches_disk        => actual_x_data%integral_caches_disk


    basis_parameter   => actual_x_data%basis_parameter
    basis_info        => actual_x_data%basis_info

    treat_lsd_in_core = general_parameter%treat_lsd_in_core

    !! initalize some counters
    neris_total = 0_int_8
    neris_incore = 0_int_8
    neris_disk = 0_int_8
    neris_onthefly = 0_int_8
    mem_eris = 0_int_8
    mem_eris_disk = 0_int_8
    mem_max_val = 0_int_8
    mem_maxvals = 0_int_8
    compression_factor = 0.0_dp
    compression_factor_disk = 0.0_dp
    nprim_ints = 0_int_8
    neris_tmp = 0_int_8
    max_val_memory = 1_int_8

    do_p_screening = screening_parameter%do_initial_p_screening
    max_set = basis_info%max_set
    CALL get_qs_env(qs_env=qs_env,&
                    atomic_kind_set=atomic_kind_set,&
                    particle_set=particle_set,&
                    error=error)
 
    natom = SIZE(particle_set,1)

    ALLOCATE(kind_of(natom),STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)

    CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set,&
                             kind_of=kind_of)

    !! precompute maximum nco and allocate scratch
    nco_max=0
    nsgf_max=0
    DO iatom=1,natom
      ikind = kind_of(iatom)
      nseta = basis_parameter(ikind)%nset
      npgfa => basis_parameter(ikind)%npgf
      la_max => basis_parameter(ikind)%lmax
      nsgfa => basis_parameter(ikind)%nsgf
      DO iset = 1, nseta
         nco_max = MAX(nco_max,npgfa(iset)*ncoset(la_max(iset)))
         nsgf_max = MAX(nsgf_max,nsgfa(iset))
      ENDDO
    ENDDO
    !! Allocate the arrays for the integrals. T1 is only used as a work array in the contraction using the DGEMM calls
    ALLOCATE(primitive_integrals(nco_max**4),STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    primitive_integrals=0.0_dp
    ALLOCATE(T1(nco_max**4),STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    T1=0.0_dp
    ALLOCATE(pbd_buf(nsgf_max**2),STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    ALLOCATE(pbc_buf(nsgf_max**2),STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    ALLOCATE(pad_buf(nsgf_max**2),STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    ALLOCATE(pac_buf(nsgf_max**2),STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    ALLOCATE(kbd_buf(nsgf_max**2),STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    ALLOCATE(kbc_buf(nsgf_max**2),STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    ALLOCATE(kad_buf(nsgf_max**2),STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    ALLOCATE(kac_buf(nsgf_max**2),STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)

    nspins = qs_env%dft_control%nspins

    ALLOCATE(max_contraction(max_set,natom),STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)

    max_contraction=0.0_dp
    max_pgf = 0
    DO jatom=1,natom
      jkind = kind_of(jatom)
      lb_max => basis_parameter(jkind)%lmax
      nsetb = basis_parameter(jkind)%nset
      npgfb => basis_parameter(jkind)%npgf
      first_sgfb => basis_parameter(jkind)%first_sgf
      sphi_b => basis_parameter(jkind)%sphi
      nsgfb => basis_parameter(jkind)%nsgf
      DO jset = 1,nsetb
        ! takes the primitive to contracted transformation into account
        ncob = npgfb(jset)*ncoset(lb_max(jset))
        sgfb = first_sgfb(1,jset)
        ! if the primitives are assumed to be all of max_val2, max_val2*p2s_b becomes
        ! the maximum value after multiplication with sphi_b
        max_contraction(jset,jatom) = MAXVAL((/(SUM(ABS(sphi_b(1:ncob,i))),i=sgfb,sgfb+nsgfb(jset)-1)/))
        max_pgf = MAX(max_pgf,npgfb(jset))
      ENDDO
    ENDDO

    N_pmat = rho%rho_ao(1)%matrix%nrow
    ! ** Reset storage counter given by MAX_MEMORY by subtracting all buffers
    ! ** Fock and density Matrices (shared)
    subtr_size_mb = 2_int_8*N_pmat*(N_pmat+1)/2
    ! ** if non restricted ==> alpha/beta spin
    IF (.NOT. treat_lsd_in_core ) THEN 
      IF( nspins == 2) subtr_size_mb = subtr_size_mb * 2_int_8
    END IF
    ! ** Initial P only MAX(alpha,beta) (shared)
    IF(do_p_screening) THEN 
      subtr_size_mb = subtr_size_mb+max_set**2*natom**2
    END IF
    ! ** primitive buffers (not shared by the threads)
    subtr_size_mb = subtr_size_mb + 2_int_8*nco_max**4*n_threads
    ! ** screening functions (shared)
    ! ** coeffs_pgf
    subtr_size_mb = subtr_size_mb + max_pgf**2*max_set**2*nkind**2
    ! ** coeffs_set
    subtr_size_mb = subtr_size_mb + max_set**2*nkind**2
    ! ** coeffs_kind
    subtr_size_mb = subtr_size_mb + nkind**2
    ! ** radii_pgf
    subtr_size_mb = subtr_size_mb + max_pgf**2*max_set**2*nkind**2

    ! ** is_assoc (shared)
    subtr_size_mb = subtr_size_mb + natom**2

    ! ** Convert into MB's
    subtr_size_mb = subtr_size_mb * 8_int_8/1024_int_8/1024_int_8

    ! ** Subtracting all these buffers from from MAX_MEMORY yields the amount
    ! ** of RAM that is left for the compressed integrals. When using threads 
    ! ** all the available memory is shared among all n_threads. i.e. the faster
    ! ** ones can steal from the slower ones

    CALL hfx_reset_memory_usage_counter(memory_parameter, subtr_size_mb, error)

    use_disk_storage = .FALSE.
    counter = 0_int_8
    do_disk_storage = memory_parameter%do_disk_storage
    IF(geometry_did_change) THEN
     memory_parameter%ram_counter = HUGE(memory_parameter%ram_counter) 
    END IF
   
    !! We do not need the containers if MAX_MEM = 0
    IF(.NOT. memory_parameter%do_all_on_the_fly) THEN 
      !! IF new md step -> reinitialize containers
      IF(geometry_did_change) THEN
        CALL hfx_init_container(6, maxval_container, memory_parameter%actual_memory_usage, .FALSE., error)
        DO i=1,64
          CALL hfx_init_container(6, integral_containers(i),memory_parameter%actual_memory_usage, .FALSE., error)
        END DO
      END IF

      !! Decompress the first cache for maxvals and integrals
      IF( .NOT. geometry_did_change) THEN
        CALL hfx_decompress_first_cache(6, maxval_cache, maxval_container, memory_parameter%actual_memory_usage, .FALSE.)
        DO i=1,64
          CALL hfx_decompress_first_cache(i, integral_caches(i), integral_containers(i), &
                                           memory_parameter%actual_memory_usage,.FALSE.)
        END DO
      END IF
    END IF
    
    !! Since the I/O routines are no thread-safe, i.e. the procedure to get the unit number, put a lock here
!$OMP CRITICAL
    !! If we do disk storage, we have to initialize the containers/caches
    IF(do_disk_storage) THEN
      !! IF new md step -> reinitialize containers
      IF(geometry_did_change) THEN
        CALL hfx_init_container(6, maxval_container_disk, memory_parameter%actual_memory_usage_disk, do_disk_storage, error)
        DO i=1,64
          CALL hfx_init_container(i, integral_containers_disk(i),memory_parameter%actual_memory_usage_disk, do_disk_storage, error)
        END DO
      END IF
      !! Decompress the first cache for maxvals and integrals
      IF( .NOT. geometry_did_change) THEN
        CALL hfx_decompress_first_cache(6, maxval_cache_disk, maxval_container_disk, &
                                         memory_parameter%actual_memory_usage_disk, .TRUE.)
        DO i=1,64
          CALL hfx_decompress_first_cache(i, integral_caches_disk(i), integral_containers_disk(i), &
                                           memory_parameter%actual_memory_usage_disk, .TRUE.)
        END DO
      END IF
    END IF
!$OMP END CRITICAL


    !! Get screening parameter
    eps_schwarz = screening_parameter%eps_schwarz
    IF( eps_schwarz < 0.0_dp) THEN
      log10_eps_schwarz = log_zero
    ELSE
      log10_eps_schwarz = LOG10(eps_schwarz) 
    END IF
    !! get storage epsilon
    eps_storage = eps_schwarz*memory_parameter%eps_storage_scaling
    cache_size = memory_parameter%cache_size


    !! If we have a hybrid functional, we may need only a fraction of exact exchange
    hf_fraction = general_parameter%fraction 
   
    !! The number of integrals that fit into the given MAX_MEMORY

    !! Parameters related to the potential 1/r, erf(wr)/r, erfc(wr/r)
    potential_parameter = actual_x_data%potential_parameter
    
    !! Variable to check if we calculate the integrals in-core or on the fly
    !! If TRUE -> on the fly
    IF(.NOT. memory_parameter%do_all_on_the_fly) THEN
      buffer_overflow = .FALSE. 
    ELSE
      buffer_overflow = .TRUE.
    END IF
    logger => cp_error_get_logger(error)
   

    maxsgf = basis_info%max_sgf
    max_am = basis_info%max_am
    !! Initialize memory used by libint
    CALL initialize_libint(private_lib, max_am, error)

    !! Helper array to map local basis function indeces to global ones 
    ALLOCATE(last_sgf_global(0:natom),STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    last_sgf_global(0)=0
    DO iatom=1,natom
      ikind = kind_of(iatom)
      last_sgf_global(iatom) = last_sgf_global(iatom-1)+ basis_parameter(ikind)%nsgf_total
    END DO
!$OMP BARRIER
!$OMP MASTER
    !! Calculate helper array that stores if a certain atomic pair is associated in the KS matrix
    IF( geometry_did_change ) THEN
      CALL get_is_assoc_atomic_block(ks_matrix(1)%matrix,basis_parameter,kind_of,&
                    actual_x_data%is_assoc_atomic_block,actual_x_data%number_of_p_entries ,para_env, error)
      shm_number_of_p_entries = actual_x_data%number_of_p_entries
      shm_is_assoc_atomic_block => actual_x_data%is_assoc_atomic_block

      !! Get occupation of KS-matrix
      ks_fully_occ = .TRUE.
      outer: DO iatom=1,natom
        DO jatom=iatom,natom
          IF(shm_is_assoc_atomic_block(jatom,iatom) == 0 ) THEN
            ks_fully_occ = .FALSE.
            EXIT outer
          END IF
        END DO
      END DO outer
        
      IF ( .NOT. ks_fully_occ) THEN
        WRITE(error_msg,'(A,A,A)') "The Kohn Sham matrix is not 100% occupied. This "//&
             "may result in uncorrect Hartree-Fock results. Try to decrease EPS_PGF_ORB "//&
             "in the QS section. "
        CALL cp_assert(.FALSE.,cp_warning_level,cp_assertion_failed,routineP,&
                       error_msg,&
                       error,&
                       only_ionode=.TRUE.)
      END IF
    END IF

    !! Let master thread get the density (avoid problems with MPI)
    !! Get the full density from all the processors
    NULLIFY(full_density)
    NULLIFY(full_density_beta)
    ALLOCATE(full_density(N_pmat*(N_pmat+1)/2),STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)

    IF( .NOT. treat_lsd_in_core .OR. nspins == 1 ) THEN
      CALL timeset(routineN//"_getP",handle_getP)
      CALL get_full_density(para_env, full_density, rho%rho_ao(ispin)%matrix, shm_number_of_p_entries,&
                            shm_is_assoc_atomic_block, natom, last_sgf_global, &
                            kind_of, basis_parameter, N_pmat, get_max_vals_spin=.FALSE., error=error)   

      IF(nspins == 2) THEN
        ALLOCATE(full_density_beta(N_pmat*(N_pmat+1)/2),STAT=stat)
        CALL get_full_density(para_env, full_density_beta, rho%rho_ao(2)%matrix, shm_number_of_p_entries,&
                              shm_is_assoc_atomic_block, natom, last_sgf_global, &
                              kind_of, basis_parameter, N_pmat, get_max_vals_spin=.FALSE., error=error)
      END IF
      CALL timestop(handle_getP)

      !! Calculate the max values of the density matrix actual_pmax stores the data from the actual density matrix
      !! and x_data%initial_p stores the same for the initial guess. The initial guess is updated only in the case of
      !! a changed geometry 
      NULLIFY(shm_initial_p) 
      IF( do_p_screening) THEN
        shm_initial_p => actual_x_data%initial_p
        IF(geometry_did_change ) THEN
          CALL update_pmax_mat(actual_x_data%initial_p, &
                               full_density,full_density_beta,&
                               natom,max_set, kind_of, basis_parameter, &
                               last_sgf_global, N_pmat, error)
        END IF
      END IF
    ELSE
      IF( do_p_screening ) THEN
        CALL timeset(routineN//"_getP",handle_getP)
        CALL get_full_density(para_env, full_density, rho%rho_ao(1)%matrix, shm_number_of_p_entries,&
                              shm_is_assoc_atomic_block, natom, last_sgf_global, &
                              kind_of, basis_parameter, N_pmat, get_max_vals_spin=.TRUE., &
                              rho_beta=rho%rho_ao(2)%matrix, error=error)   
        CALL timestop(handle_getP)

        !! Calculate the max values of the density matrix actual_pmax stores the data from the actual density matrix
        !! and x_data%initial_p stores the same for the initial guess. The initial guess is updated only in the case of
        !! a changed geometry 
        NULLIFY(shm_initial_p) 
        shm_initial_p => actual_x_data%initial_p
        IF(geometry_did_change ) THEN
          CALL update_pmax_mat(actual_x_data%initial_p, &
                               full_density,full_density_beta,&
                               natom,max_set, kind_of, basis_parameter, &
                               last_sgf_global, N_pmat, error)
        END IF
      END IF
      ! ** Now get the density(ispin)
      CALL get_full_density(para_env, full_density, rho%rho_ao(ispin)%matrix, shm_number_of_p_entries,&
                            shm_is_assoc_atomic_block, natom, last_sgf_global, &
                            kind_of, basis_parameter, N_pmat, get_max_vals_spin=.FALSE., error=error)
    END IF

    NULLIFY(full_ks, full_ks_beta)
    nrow = rho%rho_ao(1)%matrix%nrow
    ncol = rho%rho_ao(1)%matrix%ncol
    ALLOCATE(actual_x_data%full_ks_alpha(N_pmat*(N_pmat+1)/2),STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    full_ks => actual_x_data%full_ks_alpha
    full_ks = 0.0_dp
   
    IF (.NOT. treat_lsd_in_core ) THEN 
      IF(nspins==2) THEN
        ALLOCATE(actual_x_data%full_ks_beta(N_pmat*(N_pmat+1)/2),STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        full_ks_beta => actual_x_data%full_ks_beta
        full_ks_beta = 0.0_dp
      END IF
    END IF

    !! Initialize schwarz screening matrices for near field estimates and boxing screening matrices
    !! for far field estimates. The update is only performed if the geomtry of the system changed.
    !! If the system is periodic, then the corresponding routines are called and some variables 
    !! are initialized
    IF( .NOT. actual_x_data%screen_funct_is_initialized ) THEN
      CALL calc_pair_dist_radii(qs_env, private_lib, actual_x_data%potential_parameter, i_thread,&
                                actual_x_data%pair_dist_radii_pgf, max_set, max_pgf, eps_schwarz,&
                                error)
      CALL calc_screening_functions(qs_env, private_lib, actual_x_data%potential_parameter, i_thread,&
                                    actual_x_data%screen_funct_coeffs_set, actual_x_data%screen_funct_coeffs_kind, &
                                    actual_x_data%screen_funct_coeffs_pgf,actual_x_data%pair_dist_radii_pgf,&
                                    max_set, max_pgf, error)

      actual_x_data%screen_funct_is_initialized = .TRUE.
    END IF
    screen_coeffs_set   => actual_x_data%screen_funct_coeffs_set
    screen_coeffs_kind  => actual_x_data%screen_funct_coeffs_kind
    screen_coeffs_pgf   => actual_x_data%screen_funct_coeffs_pgf
    radii_pgf           => actual_x_data%pair_dist_radii_pgf

!$OMP END MASTER
!$OMP BARRIER


    !! Initialize a prefactor depending on the fraction and the number of spins
    IF(nspins == 1) THEN
      fac = 0.5_dp * hf_fraction
    ELSE
      fac = 1.0_dp * hf_fraction
    END IF

    !! Call routines that distribute the load on all processes. If we want to screen on a initial density matrix, there is
    !! an optional parameter. Of course, this is only done if the geometry did change
!$OMP BARRIER
!$OMP MASTER
    CALL timeset(routineN//"_load",handle_load)
!$OMP END MASTER
!$OMP BARRIER
    IF( geometry_did_change ) THEN
      IF( actual_x_data%b_first_load_balance_energy ) THEN
        CALL load_balance_energy(actual_x_data,eps_schwarz,particle_set,max_set,para_env,&
                                 potential_parameter,screen_coeffs_set,screen_coeffs_kind,&
                                 shm_is_assoc_atomic_block,do_periodic,load_balance_parameter, &
                                 kind_of, basis_parameter, shm_initial_p, i_thread, n_threads, &
                                 cell, do_p_screening, error)
        actual_x_data%b_first_load_balance_energy = .FALSE.
      ELSE
        CALL update_load_balance_energy(actual_x_data,para_env, &
                                        load_balance_parameter, &
                                        i_thread, n_threads, error)
      END IF
    END IF
!$OMP BARRIER
!$OMP MASTER
    CALL timestop(handle_load)
!$OMP END MASTER
!$OMP BARRIER

    !! Start caluclating integrals of the form (ab|cd) or (ij|kl)
    !! In order to do so, there is a main four-loop structre that takes into account the two symmetries
    !!
    !!   (ab|cd) = (ba|cd) = (ab|dc) = (ba|dc)
    !!
    !! by iterating in the following way
    !!
    !! DO iatom=1,natom               and       DO katom=1,natom
    !!   DO jatom=iatom,natom                     DO latom=katom,natom
    !!
    !! The third symmetry
    !!  
    !!  (ab|cd) = (cd|ab)
    !!
    !! is taken into account by the following criterion:
    !!
    !! IF(katom+latom<=iatom+jatom)  THEN
    !!   IF( ((iatom+jatom).EQ.(katom+latom) ) .AND.(katom<iatom)) CYCLE
    !!
    !! Depending on the degeneracy of an integral the exchange contribution is multiplied by a corresponding
    !! factor ( symm_fac ).
    !!
    !! If one quartet does not pass the screening we CYCLE on the outer most possible loop. Thats why we use
    !! different hierarchies of short range screening matrices.
    !!
    !! If we do a parallel run, each process owns a unique array of workloads. Here, a workload is
    !! defined as :
    !!
    !! istart, jstart, kstart, lstart, number_of_atom_quartets, initial_cpu_id
    !!
    !! This tells the process where to start the main loops and how many bunches of integrals it has to
    !! calculate. The original parallelization is a simple modulo distribution that is binned and 
    !! optimized in the load_balance routines. Since the Monte Carlo routines can swap processors, 
    !! we need to know which was the inital cpu_id.
    !! Furthermore, the indices iatom, jatom, katom, latom have to be set to istart, jstart, kstart and
    !! lstart only the first time the loop is executed. All subsequent loops have to start with one or
    !! iatom and katom respectively. Therefore, we use flags like first_j_loop etc.

!$OMP BARRIER
!$OMP MASTER
    CALL timeset(routineN//"_main",handle_main)
!$OMP END MASTER
!$OMP BARRIER

    ncpu = para_env%num_pe 
    n_processes = ncpu * n_threads

    coeffs_kind_max0=MAXVAL(screen_coeffs_kind(:,:)%x(2))
    ALLOCATE(set_list_ij((max_set*load_balance_parameter%block_size)**2),STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    ALLOCATE(set_list_kl((max_set*load_balance_parameter%block_size)**2),STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)

    my_bin_size = SIZE(actual_x_data%distribution_energy)
    ! reset timings for the new SCF round
    IF (geometry_did_change) THEN
      DO bin=1,my_bin_size
        actual_x_data%distribution_energy(bin)%time_first_scf=0.0_dp
        actual_x_data%distribution_energy(bin)%time_other_scf=0.0_dp
        actual_x_data%distribution_energy(bin)%time_forces=0.0_dp
       END DO
    ENDIF
    block_size = load_balance_parameter%block_size
    nblocks = MAX((natom+block_size-1)/block_size,1)

    DO bin=1,my_bin_size
      bintime_start=m_walltime()
      my_istart = actual_x_data%distribution_energy(bin)%istart
      my_original_cpu_id = actual_x_data%distribution_energy(bin)%original_cpu_id 
      my_current_counter = 0
      IF(actual_x_data%distribution_energy(bin)%number_of_atom_quartets == 0 .OR. &
         my_istart == -1_int_8) my_istart = nblocks **4
      atomic_blocks:  DO atom_block = my_istart,nblocks**4-1,n_processes
        latom_block = MODULO(atom_block,nblocks)+1
        tmp_block = atom_block/nblocks
        katom_block = MODULO(tmp_block,nblocks)+1
        IF(latom_block<katom_block) CYCLE
        tmp_block = tmp_block/nblocks
        jatom_block = MODULO(tmp_block,nblocks)+1
        tmp_block = tmp_block/nblocks
        iatom_block = MODULO(tmp_block,nblocks)+1
        IF(jatom_block<iatom_block) CYCLE
        my_current_counter = my_current_counter + 1
        IF(my_current_counter > actual_x_data%distribution_energy(bin)%number_of_atom_quartets) EXIT atomic_blocks

        CALL build_pair_list(list_ij,set_list_ij,(iatom_block-1)*block_size +1,MIN(iatom_block*block_size,natom), &
                             (jatom_block-1)*block_size +1, MIN(jatom_block*block_size,natom),&
                             kind_of,basis_parameter,particle_set, &
                             do_periodic,screen_coeffs_set,screen_coeffs_kind,coeffs_kind_max0,log10_eps_schwarz,cell)

        CALL build_pair_list(list_kl,set_list_kl,(katom_block-1)*block_size +1,MIN(katom_block*block_size,natom), &
                             (latom_block-1)*block_size +1, MIN(latom_block*block_size,natom),&
                             kind_of,basis_parameter,particle_set, &
                             do_periodic,screen_coeffs_set,screen_coeffs_kind,coeffs_kind_max0,log10_eps_schwarz,cell)

        DO i_list_ij=1,list_ij%n_element

          iatom=list_ij%elements(i_list_ij)%pair(1)
          jatom=list_ij%elements(i_list_ij)%pair(2)
          i_set_list_ij_start=list_ij%elements(i_list_ij)%set_bounds(1)
          i_set_list_ij_stop=list_ij%elements(i_list_ij)%set_bounds(2)
          ikind=list_ij%elements(i_list_ij)%kind_pair(1)
          jkind=list_ij%elements(i_list_ij)%kind_pair(2)
          ra=list_ij%elements(i_list_ij)%r1
          rb=list_ij%elements(i_list_ij)%r2
          B11=list_ij%elements(i_list_ij)%vec
          rab2=list_ij%elements(i_list_ij)%dist2

          la_max => basis_parameter(ikind)%lmax
          la_min => basis_parameter(ikind)%lmin
          npgfa => basis_parameter(ikind)%npgf
          nseta = basis_parameter(ikind)%nset
          zeta => basis_parameter(ikind)%zet
          nsgfa => basis_parameter(ikind)%nsgf
          first_sgfa => basis_parameter(ikind)%first_sgf
          sphi_a => basis_parameter(ikind)%sphi
          nsgfa_total = basis_parameter(ikind)%nsgf_total
          pa1 = last_sgf_global(iatom) - nsgfa_total

          lb_max => basis_parameter(jkind)%lmax
          lb_min => basis_parameter(jkind)%lmin
          npgfb => basis_parameter(jkind)%npgf
          nsetb = basis_parameter(jkind)%nset
          zetb => basis_parameter(jkind)%zet
          nsgfb => basis_parameter(jkind)%nsgf
          first_sgfb => basis_parameter(jkind)%first_sgf
          sphi_b => basis_parameter(jkind)%sphi
          nsgfb_total = basis_parameter(jkind)%nsgf_total
          pb1 = last_sgf_global(jatom) - nsgfb_total

          DO i_list_kl=1,list_kl%n_element
            katom=list_kl%elements(i_list_kl)%pair(1)
            latom=list_kl%elements(i_list_kl)%pair(2)

            IF (.NOT. (katom+latom<=iatom+jatom)) CYCLE
            IF( ((iatom+jatom).EQ.(katom+latom) ) .AND.(katom<iatom)) CYCLE

            i_set_list_kl_start=list_kl%elements(i_list_kl)%set_bounds(1)
            i_set_list_kl_stop=list_kl%elements(i_list_kl)%set_bounds(2)
            kkind=list_kl%elements(i_list_kl)%kind_pair(1)
            lkind=list_kl%elements(i_list_kl)%kind_pair(2)
            rc=list_kl%elements(i_list_kl)%r1
            rd=list_kl%elements(i_list_kl)%r2
            D11=list_kl%elements(i_list_kl)%vec
            rcd2=list_kl%elements(i_list_kl)%dist2

            IF( (screen_coeffs_kind(jkind,ikind)%x(1)*rab2+&
                 screen_coeffs_kind(jkind,ikind)%x(2) ) + &
                (screen_coeffs_kind(lkind,kkind)%x(1)*rcd2+&
                 screen_coeffs_kind(lkind,kkind)%x(2)) < log10_eps_schwarz ) CYCLE

            !! we want to be consistent with the KS matrix. If none of the atomic indices
            !! is associated cycle
            IF(.NOT. (shm_is_assoc_atomic_block(latom,iatom)>=1 .AND. &
                      shm_is_assoc_atomic_block(katom,iatom)>=1 .AND. &
                      shm_is_assoc_atomic_block(katom,jatom)>=1 .AND. &
                      shm_is_assoc_atomic_block(latom,jatom)>=1 ) ) CYCLE
  
            !! calculate symmetry_factor accordin to degeneracy of atomic quartet
            symm_fac = 0.5_dp
            IF(iatom==jatom) symm_fac = symm_fac*2.0_dp
            IF(katom==latom) symm_fac = symm_fac*2.0_dp
            IF(iatom==katom .AND. jatom==latom .AND. iatom/=jatom .AND. katom/=latom) symm_fac = symm_fac*2.0_dp
            IF(iatom==katom .AND. iatom==jatom .AND. katom==latom) symm_fac = symm_fac*2.0_dp
            symm_fac = 1.0_dp / symm_fac

            lc_max => basis_parameter(kkind)%lmax
            lc_min => basis_parameter(kkind)%lmin
            npgfc => basis_parameter(kkind)%npgf
            nsetc = basis_parameter(kkind)%nset
            zetc => basis_parameter(kkind)%zet
            nsgfc => basis_parameter(kkind)%nsgf
            first_sgfc => basis_parameter(kkind)%first_sgf
            sphi_c => basis_parameter(kkind)%sphi
            nsgfc_total = basis_parameter(kkind)%nsgf_total
            pc1 = last_sgf_global(katom) - nsgfc_total

            ld_max => basis_parameter(lkind)%lmax
            ld_min => basis_parameter(lkind)%lmin
            npgfd => basis_parameter(lkind)%npgf
            nsetd = basis_parameter(lkind)%nset
            zetd => basis_parameter(lkind)%zet
            nsgfd => basis_parameter(lkind)%nsgf
            first_sgfd => basis_parameter(lkind)%first_sgf
            sphi_d => basis_parameter(lkind)%sphi
            nsgfd_total = basis_parameter(lkind)%nsgf_total
            pd1 = last_sgf_global(latom) - nsgfd_total
            
            !! At this stage, check for memory used in compression

            IF(geometry_did_change ) THEN
              IF(.NOT.memory_parameter%do_all_on_the_fly) THEN
                ! ** We know the maximum amount of integrals that we can store per MPI-process
                ! ** Now we need to sum the current memory usage among all openMP threads
                ! ** We can just read what is currently stored on the corresponding x_data type
                ! ** If this is thread i and it tries to read the data from thread j, that is 
                ! ** currently writing that data, we just dont care, because the possible error
                ! ** is of the order of CACHE_SIZE
                mem_compression_counter = 0
                DO i=1,n_threads
                  mem_compression_counter = mem_compression_counter + &
                                            qs_env%x_data(irep,i)%memory_parameter%actual_memory_usage *&
                                            memory_parameter%cache_size
                END DO
                IF(mem_compression_counter > memory_parameter%max_compression_counter) THEN
                  buffer_overflow = .TRUE.
                  memory_parameter%ram_counter = counter
                ELSE
                  counter = counter + 1
                  buffer_overflow = .FALSE.
                END IF 
              END IF
            ELSE
              IF(.NOT.memory_parameter%do_all_on_the_fly) THEN
                IF(memory_parameter%ram_counter == counter) THEN
                  buffer_overflow = .TRUE.
                ELSE
                  counter = counter + 1
                  buffer_overflow = .FALSE.
                END IF
              END IF
            END IF

            IF( buffer_overflow .AND. do_disk_storage ) THEN
              use_disk_storage = .TRUE.
              buffer_overflow = .FALSE.
            END IF
            
            IF( use_disk_storage ) THEN
              mem_compression_counter_disk = memory_parameter%actual_memory_usage_disk *&
                                             memory_parameter%cache_size
              IF(mem_compression_counter_disk > memory_parameter%max_compression_counter_disk) THEN
                buffer_overflow = .TRUE.
                use_disk_storage = .FALSE.
              END IF
            END IF

            DO i_set_list_ij=i_set_list_ij_start, i_set_list_ij_stop
              iset=set_list_ij(i_set_list_ij)%pair(1)
              jset=set_list_ij(i_set_list_ij)%pair(2)

              ncoa = npgfa(iset)*ncoset(la_max(iset))
              sgfa = first_sgfa(1,iset)
              ncob = npgfb(jset)*ncoset(lb_max(jset))
              sgfb = first_sgfb(1,jset)
              max_val1 = screen_coeffs_set(jset,iset,jkind,ikind)%x(1)*rab2 + &
                         screen_coeffs_set(jset,iset,jkind,ikind)%x(2)
              !! Near field screening
              IF (max_val1 + (screen_coeffs_kind(lkind,kkind)%x(1)*rcd2 + &
                              screen_coeffs_kind(lkind,kkind)%x(2) ) < log10_eps_schwarz) CYCLE
              DO i_set_list_kl=i_set_list_kl_start, i_set_list_kl_stop
                kset=set_list_kl(i_set_list_kl)%pair(1)
                lset=set_list_kl(i_set_list_kl)%pair(2)

                ncoc = npgfc(kset)*ncoset(lc_max(kset))
                sgfc = first_sgfc(1,kset)
                ncod = npgfd(lset)*ncoset(ld_max(lset)) 
                sgfd = first_sgfd(1,lset)
                max_val2_set = (screen_coeffs_set(lset,kset,lkind,kkind)%x(1)*rcd2 + &
                                       screen_coeffs_set(lset,kset,lkind,kkind)%x(2) )
                max_val2 = max_val1 + max_val2_set
                                       
                !! Near field screening
                IF(max_val2<log10_eps_schwarz) CYCLE  
              
                !! get max_vals if we screen on initial density
                IF( do_p_screening) THEN
                  pmax_entry = MAX(shm_initial_p(kset,iset,katom,iatom), &
                                   shm_initial_p(lset,jset,latom,jatom), &
                                   shm_initial_p(lset,iset,latom,iatom), &
                                   shm_initial_p(kset,jset,katom,jatom))
                ELSE
                  pmax_entry = 0.0_dp
                END IF

                log10_pmax = pmax_entry
                max_val2 = max_val2 + log10_pmax
                IF(max_val2<log10_eps_schwarz) CYCLE
                pmax_entry = EXP(log10_pmax*ln_10)
                !! store current number of integrals, update total number and number of integrals in buffer
                current_counter = nsgfa(iset)*nsgfb(jset)*nsgfc(kset)*nsgfd(lset)
                IF(buffer_overflow) THEN
                  neris_onthefly = neris_onthefly + current_counter 
                END IF
  
                !! Get integrals from buffer and update Kohn-Sham matrix
                IF(.NOT.buffer_overflow .AND. .NOT.geometry_did_change) THEN
                  nints = current_counter
                  IF(.NOT. use_disk_storage) THEN
                    CALL hfx_get_single_cache_element(estimate_to_store_int, 6,&
                                             maxval_cache, maxval_container, memory_parameter%actual_memory_usage, &
                                             use_disk_storage)
                  ELSE
                     CALL hfx_get_single_cache_element(estimate_to_store_int, 6,&
                                             maxval_cache_disk, maxval_container_disk, &
                                             memory_parameter%actual_memory_usage_disk, &
                                             use_disk_storage)
                  END IF
                  spherical_estimate = SET_EXPONENT(1.0_dp,estimate_to_store_int+1)
                  IF(spherical_estimate*pmax_entry<eps_schwarz) CYCLE
                  nbits = EXPONENT(spherical_estimate*pmax_entry/eps_storage) + 1                         
                  buffer_left = nints
                  buffer_start = 1
                  DO WHILE (buffer_left > 0 )
                    buffer_size = MIN(buffer_left, cache_size)
                    IF(.NOT. use_disk_storage) THEN
                      CALL hfx_get_mult_cache_elements(primitive_integrals(buffer_start), &
                                                        buffer_size, nbits, &
                                                        integral_caches(nbits), &
                                                        integral_containers(nbits), &
                                                        eps_storage, pmax_entry, &
                                                        memory_parameter%actual_memory_usage, &
                                                        use_disk_storage, error)
                    ELSE
                      CALL hfx_get_mult_cache_elements(primitive_integrals(buffer_start), &
                                                        buffer_size, nbits, &
                                                        integral_caches_disk(nbits), &
                                                        integral_containers_disk(nbits), &
                                                        eps_storage, pmax_entry, &
                                                        memory_parameter%actual_memory_usage_disk, &
                                                        use_disk_storage, error)
                    END IF
                    buffer_left = buffer_left - buffer_size
                    buffer_start = buffer_start + buffer_size
                  END DO 
                END IF
                !! Calculate integrals if we run out of buffer or the geometry did change
                IF(geometry_did_change .OR. buffer_overflow) THEN
                  max_contraction_val =  max_contraction(iset,iatom) * &
                                         max_contraction(jset,jatom) * &
                                         max_contraction(kset,katom) * &
                                         max_contraction(lset,latom) * pmax_entry
                  tmp_R_1 => radii_pgf(:,:,jset,iset,jkind,ikind)
                  tmp_R_2 => radii_pgf(:,:,lset,kset,lkind,kkind)
                  tmp_screen_pgf1 => screen_coeffs_pgf(:,:,jset,iset,jkind,ikind)
                  tmp_screen_pgf2 => screen_coeffs_pgf(:,:,lset,kset,lkind,kkind)
                  IF( .NOT.do_periodic ) THEN 
                    CALL coulomb4(private_lib, ra, rb, rc, rd, npgfa(iset), npgfb(jset), npgfc(kset), npgfd(lset),&
                                  la_min(iset), la_max(iset), lb_min(jset), lb_max(jset),&
                                  lc_min(kset), lc_max(kset), ld_min(lset), ld_max(lset),&
                                  ncoa, ncob, ncoc, ncod,&
                                  zeta(1:npgfa(iset),iset), zetb(1:npgfb(jset),jset),&
                                  zetc(1:npgfc(kset),kset), zetd(1:npgfd(lset),lset),&
                                  primitive_integrals,&
                                  potential_parameter, eps_schwarz, max_contraction_val, &
                                  cartesian_estimate, neris_tmp, rab2, rcd2, max_val2_set,&
                                  log10_eps_schwarz, log10_pmax, tmp_R_1, tmp_R_2, tmp_screen_pgf1, tmp_screen_pgf2,&
                                  B11, D11)
                  ELSE
                    CALL coulomb4_per(private_lib, ra, rb, rc, rd, npgfa(iset), npgfb(jset), npgfc(kset), npgfd(lset), &
                                      la_min(iset), la_max(iset), lb_min(jset), lb_max(jset),&
                                      lc_min(kset), lc_max(kset), ld_min(lset), ld_max(lset),&
                                      ncoa, ncob, ncoc, ncod,&
                                      zeta(1:npgfa(iset),iset), zetb(1:npgfb(jset),jset),&
                                      zetc(1:npgfc(kset),kset), zetd(1:npgfd(lset),lset),&
                                      primitive_integrals,&
                                      potential_parameter, &
                                      actual_x_data%neighbor_cells, screen_coeffs_set(jset,iset,jkind,ikind)%x,&
                                      screen_coeffs_set(lset,kset,lkind,kkind)%x, eps_schwarz, &
                                      max_contraction_val, cartesian_estimate,  cell, neris_tmp,&
                                      log10_pmax, B11, D11, log10_eps_schwarz, &
                                      tmp_R_1, tmp_R_2, tmp_screen_pgf1, tmp_screen_pgf2,error)
                  END IF
                  nints = nsgfa(iset)*nsgfb(jset)*nsgfc(kset)*nsgfd(lset)
                  neris_total = neris_total + nints
                  nprim_ints = nprim_ints + neris_tmp
                  IF(cartesian_estimate == 0.0_dp) cartesian_estimate = TINY(cartesian_estimate)
                  estimate_to_store_int = EXPONENT(cartesian_estimate)
                  estimate_to_store_int = MAX(estimate_to_store_int,-15_int_8)
                  cartesian_estimate = SET_EXPONENT(1.0_dp,estimate_to_store_int+1)
                  IF(.NOT.buffer_overflow .AND. geometry_did_change) THEN
                    IF(cartesian_estimate < eps_schwarz ) THEN
                      IF(.NOT. use_disk_storage) THEN
                        CALL hfx_add_single_cache_element(estimate_to_store_int, 6,&
                                        maxval_cache, maxval_container, memory_parameter%actual_memory_usage, &
                                        use_disk_storage, max_val_memory, error)
                      ELSE
                        CALL hfx_add_single_cache_element(estimate_to_store_int, 6,&
                                        maxval_cache_disk, maxval_container_disk, &
                                        memory_parameter%actual_memory_usage_disk, &
                                        use_disk_storage, error=error)
                      END IF
                    END IF
                  END IF

                  IF(cartesian_estimate<eps_schwarz) CYCLE 
                  CALL dgemm("T","N",ncob*ncoc*ncod,nsgfa(iset),ncoa,&
                             1.0_dp, primitive_integrals(1),ncoa,&
                             sphi_a(1,sgfa), SIZE(sphi_a,1),&
                             0.0_dp, T1(1),ncob*ncoc*ncod)
    
                  CALL dgemm("T","N",nsgfa(iset)*ncoc*ncod,nsgfb(jset),ncob,&
                             1.0_dp, T1(1),ncob,&
                             sphi_b(1,sgfb), SIZE(sphi_b,1),&
                             0.0_dp, primitive_integrals(1),nsgfa(iset)*ncoc*ncod)
    
                  CALL dgemm("T","N",nsgfa(iset)*nsgfb(jset)*ncod,nsgfc(kset),ncoc,&
                             1.0_dp, primitive_integrals(1),ncoc,&
                             sphi_c(1,sgfc), SIZE(sphi_c,1),&
                             0.0_dp, T1(1),nsgfa(iset)*nsgfb(jset)*ncod)
    
                  CALL dgemm("T","N",nsgfa(iset)*nsgfb(jset)*nsgfc(kset),nsgfd(lset),ncod,&
                             1.0_dp, T1(1),ncod,&
                             sphi_d(1,sgfd), SIZE(sphi_d,1),&
                             0.0_dp, primitive_integrals(1),nsgfa(iset)*nsgfb(jset)*nsgfc(kset))
                  !! Compress the array for storage
                  spherical_estimate = 0.0_dp
                  DO i=1,nints
                    spherical_estimate = MAX(spherical_estimate,ABS(primitive_integrals(i)))
                  END DO
                  IF(spherical_estimate == 0.0_dp) spherical_estimate = TINY(spherical_estimate)
                  estimate_to_store_int = EXPONENT(spherical_estimate)
                  estimate_to_store_int = MAX(estimate_to_store_int, -15_int_8)
                  
                  IF(.NOT. buffer_overflow .AND. geometry_did_change) THEN
                    IF(.NOT.use_disk_storage) THEN
                      CALL hfx_add_single_cache_element(estimate_to_store_int, 6, &
                                                         maxval_cache, maxval_container, &
                                                         memory_parameter%actual_memory_usage, &
                                                         use_disk_storage, max_val_memory, error)
                    ELSE
                      CALL hfx_add_single_cache_element(estimate_to_store_int, 6, &
                                                         maxval_cache_disk, maxval_container_disk, &
                                                         memory_parameter%actual_memory_usage_disk, &
                                                         use_disk_storage, error=error)
                    END IF
                  END IF
                  spherical_estimate = SET_EXPONENT(1.0_dp,estimate_to_store_int+1)
                  IF(spherical_estimate*pmax_entry<eps_schwarz) CYCLE 
                  IF(.NOT.buffer_overflow) THEN
                    nbits = EXPONENT(spherical_estimate*pmax_entry/eps_storage) + 1
                    buffer_left = nints
                    buffer_start = 1
                    IF( .NOT. use_disk_storage) THEN
                      neris_incore = neris_incore + nints
                    ELSE
                      neris_disk = neris_disk + nints
                    END IF
                    DO WHILE (buffer_left > 0 )
                      buffer_size = MIN(buffer_left, CACHE_SIZE)
                      IF(.NOT. use_disk_storage) THEN
                        CALL hfx_add_mult_cache_elements(primitive_integrals(buffer_start),&
                                                          buffer_size, nbits, &
                                                          integral_caches(nbits), &
                                                          integral_containers(nbits), &
                                                          eps_storage, pmax_entry, &
                                                          memory_parameter%actual_memory_usage, &
                                                          use_disk_storage, error)
                      ELSE
                        CALL hfx_add_mult_cache_elements(primitive_integrals(buffer_start),&
                                                          buffer_size, nbits, &
                                                          integral_caches_disk(nbits), &
                                                          integral_containers_disk(nbits), &
                                                          eps_storage, pmax_entry, &
                                                          memory_parameter%actual_memory_usage_disk, &
                                                          use_disk_storage, error)
                      END IF
                      buffer_left = buffer_left - buffer_size
                      buffer_start = buffer_start + buffer_size
                    END DO
                  ELSE
                    !! In order to be consistent with in-core part, round all the eris wrt. eps_schwarz
                    DO i=1,nints
                      primitive_integrals(i) = primitive_integrals(i)*pmax_entry
                      IF(ABS(primitive_integrals(i)) > eps_storage) THEN
                        primitive_integrals(i) = ANINT(primitive_integrals(i)/eps_storage, dp)*eps_storage/pmax_entry
                      ELSE
                        primitive_integrals(i) = 0.0_dp
                      END IF
                    END DO
                  END IF
                END IF
                !! Update Kohn-Sham matrix
                CALL update_fock_matrix(pa1, pb1, pc1, pd1, &
                                        nsgfa(iset), nsgfb(jset), nsgfc(kset), nsgfd(lset), &
                                        sgfa, sgfb, sgfc, sgfd, fac, symm_fac, full_density, full_ks, &
                                        primitive_integrals, pbd_buf, pbc_buf, pad_buf, pac_buf, kbd_buf,&
                                        kbc_buf, kad_buf, kac_buf, N_pmat)
                IF( .NOT. treat_lsd_in_core ) THEN
                  IF( nspins == 2 ) THEN
                    CALL update_fock_matrix(pa1, pb1, pc1, pd1, &
                                            nsgfa(iset), nsgfb(jset), nsgfc(kset), nsgfd(lset), &
                                            sgfa, sgfb, sgfc, sgfd, fac, symm_fac, full_density_beta, full_ks_beta, &
                                            primitive_integrals, pbd_buf, pbc_buf, pad_buf, pac_buf, kbd_buf,&
                                            kbc_buf, kad_buf, kac_buf, N_pmat)
                  END IF
                END IF
              END DO ! i_set_list_kl
            END DO ! i_set_list_ij
            IF( do_disk_storage ) THEN
              buffer_overflow = .TRUE.
            END IF
          END DO !i_list_ij
        END DO !i_list_kl
      END DO atomic_blocks
      bintime_stop = m_walltime()
      IF (geometry_did_change) THEN
        actual_x_data%distribution_energy(bin)%time_first_scf=bintime_stop-bintime_start 
      ELSE
        actual_x_data%distribution_energy(bin)%time_other_scf= &
            actual_x_data%distribution_energy(bin)%time_other_scf+bintime_stop-bintime_start 
      ENDIF
    END DO !bin

!$OMP BARRIER
!$OMP MASTER
    memsize_after=m_memory()
!$OMP END MASTER
!$OMP BARRIER

    DEALLOCATE(primitive_integrals,STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    DEALLOCATE(T1,STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
!$OMP BARRIER
    !! Get some number about ERIS
!$OMP ATOMIC
    shm_neris_total = shm_neris_total + neris_total
!$OMP ATOMIC
    shm_neris_onthefly = shm_neris_onthefly + neris_onthefly
!$OMP ATOMIC
    shm_nprim_ints = shm_nprim_ints + nprim_ints

    storage_counter_integrals = memory_parameter%actual_memory_usage * &
                                memory_parameter%cache_size
    stor_count_int_disk = memory_parameter%actual_memory_usage_disk * &
                          memory_parameter%cache_size
    stor_count_max_val = max_val_memory * memory_parameter%cache_size
!$OMP ATOMIC
    shm_storage_counter_integrals = shm_storage_counter_integrals + storage_counter_integrals
!$OMP ATOMIC
    shm_stor_count_int_disk =  shm_stor_count_int_disk + stor_count_int_disk
!$OMP ATOMIC
    shm_neris_incore = shm_neris_incore + neris_incore
!$OMP ATOMIC
    shm_neris_disk = shm_neris_disk + neris_disk
!$OMP ATOMIC
    shm_stor_count_max_val = shm_stor_count_max_val + stor_count_max_val
!$OMP BARRIER

!$OMP MASTER   
    !! Calculate the exchange energies from the Kohn-Sham matrix. Before we can go on, we have to symmetrize.
    ene_x_aa = 0.0_dp
    ene_x_bb = 0.0_dp
   
    mb_size_p = N_pmat*(N_pmat+1)/2 /1024/128
    mb_size_f = N_pmat*(N_pmat+1)/2 /1024/128
    IF (.NOT. treat_lsd_in_core ) THEN
      IF( nspins == 2) THEN
        mb_size_f = mb_size_f * 2
        mb_size_p = mb_size_p * 2
      END IF
    END IF
    !! size of primitive_integrals and T1 buffer (not shared)
    mb_size_buffers = 2*INT(nco_max,int_8)**4 * n_threads
    !! size of screening functions (shared)
    mb_size_buffers = mb_size_buffers + max_pgf**2*max_set**2*nkind**2 &
                                      + max_set**2*nkind**2 &
                                      + nkind**2 &
                                      + max_pgf**2*max_set**2*nkind**2
    !! is_assoc (shared)
    mb_size_buffers = mb_size_buffers+natom**2
    !! mb
    mb_size_buffers = mb_size_buffers /1024/128

    CALL timestop(handle_main)
    
    IF( distribute_fock_matrix) THEN
      !! Distribute the current KS-matrix to all the processes
      CALL timeset(routineN//"_dist_KS",handle_dist_ks)
      CALL distribute_ks_matrix(para_env, full_ks, ks_matrix(ispin)%matrix, shm_number_of_p_entries, &
                                shm_is_assoc_atomic_block,&
                                natom, last_sgf_global, kind_of, basis_parameter, &
                                N_pmat, off_diag_fac=0.5_dp,error=error)

      NULLIFY(full_ks)
      DEALLOCATE(actual_x_data%full_ks_alpha, STAT=stat)
      CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
      IF( .NOT. treat_lsd_in_core ) THEN
        IF(nspins == 2) THEN
          CALL distribute_ks_matrix(para_env, full_ks_beta, ks_matrix(2)%matrix, shm_number_of_p_entries, &
                                    shm_is_assoc_atomic_block,&
                                    natom, last_sgf_global, kind_of, basis_parameter, &
                                    N_pmat, off_diag_fac=0.5_dp, error=error)
          NULLIFY(full_ks_beta)
          DEALLOCATE(actual_x_data%full_ks_beta, STAT=stat)
          CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
        END IF
      END IF
      CALL timestop(handle_dist_ks)
    END IF


    IF( distribute_fock_matrix) THEN
      !! ** Calculate the exchange energy
      CALL cp_sm_sm_trace(ks_matrix(ispin)%matrix, rho%rho_ao(ispin)%matrix, &
                          ene_x_aa, para_env, error)
      IF( nspins == 2 .AND. .NOT. treat_lsd_in_core ) THEN
        CALL cp_sm_sm_trace(ks_matrix(2)%matrix, rho%rho_ao(2)%matrix, &
                            ene_x_bb, para_env, error)
      END IF
  
      !! Update energy type
      energy%ex = 0.5_dp*(ene_x_aa+ene_x_bb) + energy%ex
    ELSE
      DO pa=1,N_pmat
        DO pd=pa,N_pmat
          IF(pa/=pd) THEN
            full_ks(get_1d_idx(pa,pd,N_pmat)) = full_ks(get_1d_idx(pa,pd,N_pmat)) * 0.5_dp
            ene_x_aa = ene_x_aa + 2.0_dp * full_density(get_1d_idx(pa,pd,N_pmat))*full_ks(get_1d_idx(pa,pd,N_pmat))
          ELSE
            ene_x_aa = ene_x_aa + full_density(get_1d_idx(pa,pd,N_pmat))*full_ks(get_1d_idx(pa,pd,N_pmat))
          END IF
        END DO
      END DO
      IF( nspins == 2 ) THEN
        DO pa=1,N_pmat
          DO pd=pa,N_pmat
            IF(pa/=pd) THEN
              full_ks_beta(get_1d_idx(pa,pd,N_pmat)) = full_ks_beta(get_1d_idx(pa,pd,N_pmat)) * 0.5_dp
              ene_x_bb = ene_x_bb + 2.0_dp * full_density_beta(get_1d_idx(pa,pd,N_pmat))*full_ks_beta(get_1d_idx(pa,pd,N_pmat))
            ELSE
              ene_x_bb = ene_x_bb + full_density_beta(get_1d_idx(pa,pd,N_pmat))*full_ks_beta(get_1d_idx(pa,pd,N_pmat))
            END IF
          END DO
        END DO
      END IF
      CALL mp_sum(ene_x_aa, para_env%group)
      IF(nspins==2) CALL mp_sum(ene_x_bb,para_env%group)
      energy%ex = 0.5_dp*(ene_x_aa+ene_x_bb)
    END IF

    !! Print some memeory information if this is the first step
    IF( geometry_did_change ) THEN
      tmp_i8(1:8)=(/shm_storage_counter_integrals,shm_neris_onthefly,shm_neris_incore,shm_neris_disk, &
                    shm_neris_total,shm_stor_count_int_disk,shm_nprim_ints,shm_stor_count_max_val/)
      CALL mp_sum_i8(tmp_i8,para_env%group)
      shm_storage_counter_integrals=tmp_i8(1)
      shm_neris_onthefly=tmp_i8(2)
      shm_neris_incore=tmp_i8(3)
      shm_neris_disk=tmp_i8(4)
      shm_neris_total=tmp_i8(5)
      shm_stor_count_int_disk=tmp_i8(6)
      shm_nprim_ints=tmp_i8(7)
      shm_stor_count_max_val=tmp_i8(8)
      CALL mp_max_i8(memsize_after, para_env%group)
      mem_eris = (shm_storage_counter_integrals+128*1024-1)/1024/128
      compression_factor = REAL(shm_neris_incore,dp)/REAL(shm_storage_counter_integrals,dp)
      mem_eris_disk = (shm_stor_count_int_disk+128*1024-1)/1024/128
      compression_factor_disk = REAL(shm_neris_disk,dp)/REAL(shm_stor_count_int_disk,dp)
      mem_max_val = (shm_stor_count_max_val+128*1024-1)/1024/128
     
      IF( shm_neris_incore == 0 ) THEN 
        mem_eris = 0
        compression_factor = 0.0_dp
      END IF
      IF( shm_neris_disk == 0 ) THEN
        mem_eris_disk = 0
        compression_factor_disk = 0.0_dp
      END IF
        
      iw = cp_print_key_unit_nr(logger,hfx_section,"HF_INFO",&
           extension=".scfLog",error=error)
      IF (iw>0) THEN
        WRITE (UNIT=iw,FMT="((T3,A,T60,I20))")&
                 "HFX_MEM_INFO| Number of cart. primitive ERI's calculated:      ",shm_nprim_ints

        WRITE (UNIT=iw,FMT="((T3,A,T60,I20))")&
                 "HFX_MEM_INFO| Number of sph. ERI's calculated:           ",shm_neris_total

        WRITE (UNIT=iw,FMT="((T3,A,T60,I20))")&
                 "HFX_MEM_INFO| Number of sph. ERI's stored in-core:        ",shm_neris_incore

        WRITE (UNIT=iw,FMT="((T3,A,T60,I20))")&
                 "HFX_MEM_INFO| Number of sph. ERI's stored on disk:        ",shm_neris_disk
  
        WRITE (UNIT=iw,FMT="((T3,A,T60,I20))")&
                 "HFX_MEM_INFO| Number of sph. ERI's calculated on the fly: ",shm_neris_onthefly
  
        WRITE (UNIT=iw,FMT="((T3,A,T60,I20))")&
                 "HFX_MEM_INFO| Total memory consumption ERI's RAM [MB's]:            ",mem_eris
        
        WRITE (UNIT=iw,FMT="((T3,A,T60,I20))")&
                 "HFX_MEM_INFO| Whereof max-vals [MB's]:            ",mem_max_val
      
        WRITE (UNIT=iw,FMT="((T3,A,T60,F20.2))")&
                 "HFX_MEM_INFO| Total compression factor ERI's RAM:                  ",compression_factor
             
        WRITE (UNIT=iw,FMT="((T3,A,T60,I20))")&
                 "HFX_MEM_INFO| Total memory consumption ERI's disk [MB's]:       ",mem_eris_disk

        WRITE (UNIT=iw,FMT="((T3,A,T60,F20.2))")&
                 "HFX_MEM_INFO| Total compression factor ERI's disk:             ",compression_factor_disk

        WRITE (UNIT=iw,FMT="((T3,A,T60,I20))")&
                 "HFX_MEM_INFO| Size of density/Fock matrix [MB's]:             ", 2_int_8*mb_size_p

        IF( do_periodic) THEN
          WRITE (UNIT=iw,FMT="((T3,A,T60,I20))")&
                 "HFX_MEM_INFO| Size of buffers [MB's]:             ", mb_size_buffers
          WRITE (UNIT=iw,FMT="((T3,A,T60,I20))")&
                 "HFX_MEM_INFO| Number of periodic image cells considered: ", SIZE(actual_x_data%neighbor_cells)
        ELSE
           WRITE (UNIT=iw,FMT="((T3,A,T60,I20))")&
                 "HFX_MEM_INFO| Size of buffers [MB's]:             ", mb_size_buffers
        END IF
        WRITE (UNIT=iw,FMT="((T3,A,T60,I20),/)")&
                 "HFX_MEM_INFO| Est. max. program size after HFX  [MB's]:", memsize_after/(1024*1024)

      END IF
  
      CALL cp_print_key_finished_output(iw,logger,hfx_section,&
                                      "HF_INFO", error=error)
    END IF
!$OMP END MASTER

    !! flush caches if the geometry changed 
    IF(geometry_did_change ) THEN
      CALL hfx_flush_last_cache(6, maxval_cache, maxval_container, memory_parameter%actual_memory_usage, .FALSE., error)
      DO i=1,64
        CALL hfx_flush_last_cache(i, integral_caches(i), integral_containers(i), &
                                   memory_parameter%actual_memory_usage, .FALSE., error)
      END DO
    END IF
    !! reset all caches except we calculate all on the fly
    IF(.NOT.memory_parameter%do_all_on_the_fly) THEN
      CALL hfx_reset_cache_and_container(maxval_cache, maxval_container, memory_parameter%actual_memory_usage, .FALSE.)
      DO i=1,64
        CALL hfx_reset_cache_and_container(integral_caches(i), integral_containers(i), memory_parameter%actual_memory_usage, &
                                            .FALSE.)
      END DO
    END IF

    !! Since the I/O routines are no thread-safe, i.e. the procedure to get the unit number, put a lock here
!$OMP CRITICAL
    IF(do_disk_storage ) THEN
      !! flush caches if the geometry changed 
      IF(geometry_did_change ) THEN
        CALL hfx_flush_last_cache(6, maxval_cache_disk, maxval_container_disk, &
                                   memory_parameter%actual_memory_usage_disk, .TRUE., error)
        DO i=1,64
          CALL hfx_flush_last_cache(i, integral_caches_disk(i), integral_containers_disk(i), &
                                     memory_parameter%actual_memory_usage_disk, .TRUE., error)
        END DO
      END IF
      !! reset all caches except we calculate all on the fly
      CALL hfx_reset_cache_and_container(maxval_cache_disk, maxval_container_disk, memory_parameter%actual_memory_usage_disk, &
                                          do_disk_storage)
      DO i=1,64
        CALL hfx_reset_cache_and_container(integral_caches_disk(i), integral_containers_disk(i),&
                                            memory_parameter%actual_memory_usage_disk, do_disk_storage)
      END DO
    END IF   
!$OMP END CRITICAL
!$OMP BARRIER
    !! Clean up
    DEALLOCATE(last_sgf_global,STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
!$OMP MASTER
    DEALLOCATE(full_density,STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    IF( .NOT. treat_lsd_in_core ) THEN
      IF(nspins==2) THEN
        DEALLOCATE(full_density_beta,STAT=stat)
        CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
       END IF
    END IF
!$OMP END MASTER
    DEALLOCATE(pbd_buf, pbc_buf, pad_buf, pac_buf, STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    DEALLOCATE(kbd_buf,kbc_buf,kad_buf,kac_buf, STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    DEALLOCATE(set_list_ij, set_list_kl, STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)

    DEALLOCATE(max_contraction, kind_of, STAT=stat)
    CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
    !
    ! this wraps to free_libint, but currently causes a segfault
    ! as a result, we don't call it, but some memory remains allocated
    !  
    ! CALL terminate_libint(lib(i_thread))
    
!$OMP BARRIER
!$OMP END PARALLEL
    CALL timestop(handle)
  END SUBROUTINE integrate_four_center

! *****************************************************************************
!> \brief calculates two-electron integrals of a quartet/shell using the library 
!>      lib_int
!> \param r position
!> \param z zeta
!> \param l angular momentum
!> \param npgf number of primitive cartesian gaussian in actual shell
!> \param primitive_integrals array of primitive_integrals
!> \param potential_parameter contains info for libint 
!> \param eps_schwarz threshold
!> \param max_contraction_val maximum multiplication factor for cart -> sph
!> \param cartesian_estimate maximum calculated integral value
!> \param i_thread current thread ID
!> \param neris_tmp counter for calculated cart integrals
!> \param rab2,rcd2 (ab| (cd| distances
!> \param max_val2_set schwarz estimate for the current set product
!> \param log10_eps_schwarz log of threshold
!> \param log10_pmax logarithm of initial p matrix max element
!> \param R1_pgf, R2_pgf coefficients for radii of product distribution function
!> \param pgf1, pgf2 schwarz coefficients pgf basid
!> \param B11, D11 ra-rb and rc-rd
!> \param error variable to control error logging, stopping,...
!>        see module cp_error_handling
!> \par History
!>      11.2006 created [Manuel Guidon]
!>      02.2009 completely rewritten screening part [Manuel Guidon]
!> \author Manuel Guidon
! *****************************************************************************
  SUBROUTINE coulomb4(lib, ra, rb, rc, rd, npgfa, npgfb, npgfc, npgfd,&
                      la_min, la_max, lb_min, lb_max,&
                      lc_min, lc_max, ld_min, ld_max,&
                      ncoa, ncob, ncoc, ncod,&
                      zeta, zetb, zetc, zetd,&
                      primitive_integrals,&
                      potential_parameter, eps_schwarz, max_contraction_val, &
                      cartesian_estimate, neris_tmp, rab2, rcd2, max_val2_set,&
                      log10_eps_schwarz, log10_pmax,R1_pgf, R2_pgf, pgf1, pgf2,B11,D11)
    TYPE(lib_int)                            :: lib
    REAL(dp), INTENT(IN)                     :: ra(3), rb(3), rc(3), rd(3)
    INTEGER, INTENT(IN) :: npgfa, npgfb, npgfc, npgfd, la_min, la_max, &
      lb_min, lb_max, lc_min, lc_max, ld_min, ld_max, ncoa, ncob, ncoc, ncod
    REAL(dp), DIMENSION(1:npgfa), INTENT(IN) :: zeta
    REAL(dp), DIMENSION(1:npgfb), INTENT(IN) :: zetb
    REAL(dp), DIMENSION(1:npgfc), INTENT(IN) :: zetc
    REAL(dp), DIMENSION(1:npgfd), INTENT(IN) :: zetd
    REAL(dp), &
      DIMENSION(ncoa, ncob, ncoc, ncod)      :: primitive_integrals
    TYPE(hfx_potential_type)                 :: potential_parameter
    REAL(dp), INTENT(IN)                     :: eps_schwarz, &
                                                max_contraction_val
    REAL(dp), INTENT(INOUT)                  :: cartesian_estimate
    INTEGER(int_8)                           :: neris_tmp
    REAL(dp)                                 :: rab2, rcd2, max_val2_set, &
                                                log10_eps_schwarz, log10_pmax
    TYPE(hfx_screen_coeff_type), &
      DIMENSION(:, :), POINTER               :: R1_pgf, R2_pgf, pgf1, pgf2
    REAL(dp), INTENT(IN)                     :: B11(3), D11(3)

    INTEGER :: ipgf, jpgf, kpgf, la, lb, lc, ld, lpgf, offset_a, offset_a1, &
      offset_b, offset_b1, offset_c, offset_c1, offset_d, offset_d1
    REAL(dp) :: Eta, EtaInv, P(3), pgf_max_1, pgf_max_2, Q(3), R1, R2, Rho, &
      RhoInv, rpq2, S1234, S1234a, tmp_max, W(3), Zeta1, Zeta_A, Zeta_B, &
      Zeta_C, Zeta_D, ZetaInv, ZetapEtaInv

    neris_tmp = 0
    primitive_integrals = 0.0_dp
    cartesian_estimate = 0.0_dp
    DO ipgf = 1,npgfa
      offset_a1 = (ipgf-1)*ncoset(la_max)
      Zeta_A = zeta(ipgf)
      DO jpgf = 1,npgfb
        pgf_max_1 = pgf1(jpgf,ipgf)%x(1)*rab2+pgf1(jpgf,ipgf)%x(2)
        IF( pgf_max_1 + max_val2_set + log10_pmax < log10_eps_schwarz) CYCLE
        R1 = MAX(0.0_dp, R1_pgf(jpgf,ipgf)%x(1)*rab2 + R1_pgf(jpgf,ipgf)%x(2))
        Zeta_B = zetb(jpgf)
        Zeta1 = Zeta_A + Zeta_B
        ZetaInv = 1.0_dp/Zeta1
        S1234a = (-Zeta_A*Zeta_B*ZetaInv*rab2)
        P = (Zeta_A*ra + Zeta_B*rb)*ZetaInv
        offset_b1 = (jpgf-1)*ncoset(lb_max)
        DO kpgf = 1,npgfc
          offset_c1 = (kpgf-1)*ncoset(lc_max)
          Zeta_C = zetc(kpgf)
          DO lpgf = 1,npgfd
            pgf_max_2 = pgf2(lpgf,kpgf)%x(1)*rcd2+pgf2(lpgf,kpgf)%x(2)
            IF (pgf_max_1 + pgf_max_2 + log10_pmax < log10_eps_schwarz ) CYCLE
            R2 = MAX(0.0_dp, R2_pgf(lpgf,kpgf)%x(1)*rcd2 + R2_pgf(lpgf,kpgf)%x(2))
            Zeta_D = zetd(lpgf)
            Eta  =  Zeta_C + Zeta_D
            EtaInv = 1.0_dp/Eta
            ZetapEtaInv = Zeta1+Eta
            ZetapEtaInv = 1.0_dp/ZetapEtaInv
            Rho = Zeta1*Eta*ZetapEtaInv
            RhoInv = 1.0_dp/Rho
            S1234 = EXP(S1234a-Zeta_C*Zeta_D*EtaInv*rcd2)
            Q = (Zeta_C*rc + Zeta_D*rd)*EtaInv
            rpq2 = (P(1)-Q(1))**2+(P(2)-Q(2))**2+(P(3)-Q(3))**2
            W = (Zeta1*P+Eta*Q)*ZetapEtaInv
            offset_d1 = (lpgf-1)*ncoset(ld_max)
            DO la = la_min,la_max
              offset_a = offset_a1 + ncoset(la-1)
              DO lb = lb_min, lb_max
                offset_b = offset_b1 + ncoset(lb-1)
                DO lc = lc_min, lc_max
                  offset_c = offset_c1 + ncoset(lc-1)
                  DO ld = ld_min, ld_max
                    offset_d = offset_d1 + ncoset(ld-1)
                    !Build primitives
                    tmp_max = 0.0_dp
                    CALL evaluate_eri(lib, ra, rb, rc, rd,&
                                      zeta(ipgf), zetb(jpgf), zetc(kpgf), zetd(lpgf),&
                                      la, lb, lc ,ld,&
                                      ncoa,ncob,ncoc,ncod,&
                                      offset_a,offset_b,offset_c,offset_d, &
                                      primitive_integrals, &
                                      potential_parameter, max_contraction_val, tmp_max, eps_schwarz,&
                                      neris_tmp,Zeta1,ZetaInv,Eta,EtaInv,ZetapEtaInv,Rho,RhoInv,S1234,&
                                      P,Q,W,rpq2,B11,D11,R1,R2)
                    cartesian_estimate = MAX(cartesian_estimate, tmp_max)
                  END DO !la
                END DO !ipgf
              END DO !lb
            END DO !jpgf
          END DO !lc
        END DO !kpgf
      END DO !ld
    END DO !lpgf
  END SUBROUTINE coulomb4
! *****************************************************************************
!> \brief calculates two-electron integrals of a quartet/shell using the library 
!>      lib_int in the periodic case
!> \param r position
!> \param z zeta
!> \param l angular momentum
!> \param npgf number of primitive cartesian gaussian in actual shell
!> \param primitive_integrals array of primitive_integrals
!> \param potential_parameter contains info for libint 
!> \param neighbor_cells Periodic images
!> \param screen1, screen2 set based coefficients for near field screening
!> \param eps_schwarz threshold
!> \param max_contraction_val maximum multiplication factor for cart -> sph
!> \param cartesian_estimate maximum calculated integral value
!> \param i_thread current thread ID
!> \param cell cell
!> \param neris_tmp counter for calculated cart integrals
!> \param log10_pmax logarithm of initial p matrix max element
!> \param B11, D11 ra-rb, rc-rd after first MIC
!> \param log10_eps_schwarz log of threshold
!> \param R1_pgf, R2_pgf coefficients for radii of product distribution function
!> \param pgf1, pgf2 schwarz coefficients pgf basid
!> \param error variable to control error logging, stopping,...
!>        see module cp_error_handling
!> \par History
!>      11.2006 created [Manuel Guidon]
!>      02.2009 completely rewritten screening part [Manuel Guidon]
!> \author Manuel Guidon
! *****************************************************************************
  SUBROUTINE coulomb4_per(lib, ra, rb, rc, rd, npgfa, npgfb, npgfc, npgfd,&
                          la_min, la_max, lb_min, lb_max,&
                          lc_min, lc_max, ld_min, ld_max,&
                          ncoa, ncob, ncoc, ncod,&
                          zeta, zetb, zetc, zetd,&
                          primitive_integrals,&
                          potential_parameter, neighbor_cells, &
                          screen1, screen2, eps_schwarz, max_contraction_val, &
                          cart_estimate, cell, neris_tmp, log10_pmax,& 
                          B11,D11,log10_eps_schwarz,R1_pgf, R2_pgf, pgf1, pgf2, error)
   
    TYPE(lib_int)                            :: lib
    REAL(dp), INTENT(IN)                     :: ra(3), rb(3), rc(3), rd(3)
    INTEGER, INTENT(IN) :: npgfa, npgfb, npgfc, npgfd, la_min, la_max, &
      lb_min, lb_max, lc_min, lc_max, ld_min, ld_max, ncoa, ncob, ncoc, ncod
    REAL(dp), DIMENSION(1:npgfa)             :: zeta
    REAL(dp), DIMENSION(1:npgfb)             :: zetb
    REAL(dp), DIMENSION(1:npgfc)             :: zetc
    REAL(dp), DIMENSION(1:npgfd)             :: zetd
    REAL(dp), &
      DIMENSION(ncoa, ncob, ncoc, ncod)      :: primitive_integrals
    TYPE(hfx_potential_type)                 :: potential_parameter
    TYPE(hfx_cell_type), DIMENSION(:), &
      POINTER                                :: neighbor_cells
    REAL(dp), INTENT(IN)                     :: screen1(2), screen2(2), &
                                                eps_schwarz, &
                                                max_contraction_val
    REAL(dp), INTENT(INOUT)                  :: cart_estimate
    TYPE(cell_type), POINTER                 :: cell
    INTEGER(int_8)                           :: neris_tmp
    REAL(dp), INTENT(IN)                     :: log10_pmax, B11(3), D11(3), &
                                                log10_eps_schwarz
    TYPE(hfx_screen_coeff_type), &
      DIMENSION(:, :), POINTER               :: R1_pgf, R2_pgf, pgf1, pgf2
    TYPE(cp_error_type), INTENT(inout)       :: error

    CHARACTER(LEN=*), PARAMETER :: routineN = 'coulomb4_per', &
      routineP = moduleN//':'//routineN

    INTEGER :: i, ipgf, j, jpgf, k, kpgf, la, lb, lc, ld, lpgf, offset_a, &
      offset_a1, offset_b, offset_b1, offset_c, offset_c1, offset_d, offset_d1
    LOGICAL                                  :: failure
    REAL(dp) :: AB(3), C11(3), CD(3), Eta, EtaInv, im_B(3), im_D(3), &
      max_val1, max_val2, P(3), pgf_max_1, pgf_max_2, PQ(3), Q(3), R1, R2, &
      rab2, rcd2, Rho, RhoInv, rpq2, S1234, S1234a, shift(3), temp(3), &
      temp_CC(3), temp_DD(3), tmp_D(3), tmp_max, W(3), Zeta1, Zeta_A, Zeta_B, &
      Zeta_C, Zeta_D, ZetaInv, ZetapEtaInv

    failure = .FALSE.
    cart_estimate = 0.0_dp
    neris_tmp = 0

    primitive_integrals = 0.0_dp
    DO i = 1,SIZE(neighbor_cells)
      im_B = B11 + neighbor_cells(i)%cell_r(:)
      AB = ra-im_B
      rab2 = AB(1)**2+AB(2)**2+AB(3)**2
      max_val1 = screen1(1)*rab2+screen1(2)
      IF( max_val1 + screen2(2) + log10_pmax < log10_eps_schwarz ) CYCLE
      DO j = 1,SIZE(neighbor_cells)
        im_D = D11 + neighbor_cells(j)%cell_r(:)
        rcd2 = (im_D(1)-rc(1))**2+(im_D(2)-rc(2))**2+(im_D(3)-rc(3))**2
        max_val2 = screen2(1)*rcd2+screen2(2)
        IF( max_val1 + max_val2 + log10_pmax < log10_eps_schwarz ) CYCLE
        DO ipgf = 1,npgfa
          Zeta_A = zeta(ipgf)
          offset_a1 = (ipgf-1)*ncoset(la_max)
          DO jpgf = 1,npgfb
            pgf_max_1 = pgf1(jpgf,ipgf)%x(1)*rab2+pgf1(jpgf,ipgf)%x(2)
            IF( pgf_max_1 + max_val2 + log10_pmax < log10_eps_schwarz) CYCLE
            R1 = MAX(0.0_dp,R1_pgf(jpgf,ipgf)%x(1)*rab2 + R1_pgf(jpgf,ipgf)%x(2))
            Zeta_B = zetb(jpgf)
            offset_b1 = (jpgf-1)*ncoset(lb_max)
            Zeta1 = Zeta_A + Zeta_B
            ZetaInv = 1.0_dp/Zeta1
            S1234a = (-Zeta_A*Zeta_B*ZetaInv*rab2)
            P = (Zeta_A*ra + Zeta_B*im_B)*ZetaInv
            DO kpgf = 1,npgfc
              Zeta_C = zetc(kpgf)
              offset_c1 = (kpgf-1)*ncoset(lc_max)
              DO lpgf = 1,npgfd
                pgf_max_2 = pgf2(lpgf,kpgf)%x(1)*rcd2+pgf2(lpgf,kpgf)%x(2)
                IF (pgf_max_1 + pgf_max_2 + log10_pmax < log10_eps_schwarz ) CYCLE
                R2 = MAX(0.0_dp, R2_pgf(lpgf,kpgf)%x(1)*rcd2 + R2_pgf(lpgf,kpgf)%x(2))
                offset_d1 = (lpgf-1)*ncoset(ld_max)
                Zeta_D = zetd(lpgf)
                Eta  =  Zeta_C + Zeta_D
                EtaInv = 1.0_dp/Eta
                ZetapEtaInv = Zeta1+Eta
                ZetapEtaInv = 1.0_dp/ZetapEtaInv
                Rho = Zeta1*Eta*ZetapEtaInv
                RhoInv = 1.0_dp/Rho
                S1234 = EXP(S1234a-Zeta_C*Zeta_D*EtaInv*rcd2)
                Q = (Zeta_C*rc + Zeta_D*im_D)*EtaInv
                temp = P-Q
                PQ = pbc(temp,cell)
                shift = -PQ + temp
                temp_CC = rc+shift
                temp_DD = im_D+shift
                DO k = 1,SIZE(neighbor_cells)
                  C11 = temp_CC + neighbor_cells(k)%cell_r(:)
                  tmp_D = temp_DD + neighbor_cells(k)%cell_r(:)
                  Q = (Zeta_C*C11 + Zeta_D*tmp_D)*EtaInv
                  W = (Zeta1*P+Eta*Q)*ZetapEtaInv
                  CD = C11-tmp_D
                  rpq2 = (P(1)-Q(1))**2+(P(2)-Q(2))**2+(P(3)-Q(3))**2
                  DO la = la_min,la_max
                    offset_a = offset_a1 + ncoset(la-1)
                    DO lb = lb_min, lb_max
                      offset_b = offset_b1 + ncoset(lb-1)
                      DO lc = lc_min, lc_max
                        offset_c = offset_c1 + ncoset(lc-1)
                        DO ld = ld_min, ld_max
                          offset_d = offset_d1 + ncoset(ld-1)
                          !Build primitives
                          tmp_max = 0.0_dp
                          CALL evaluate_eri(lib, ra, im_B, C11, tmp_D,&
                                            zeta(ipgf), zetb(jpgf), zetc(kpgf), zetd(lpgf),&
                                            la, lb, lc ,ld,&
                                            ncoa,ncob,ncoc,ncod,&
                                            offset_a,offset_b,offset_c,offset_d, &
                                            primitive_integrals, &
                                            potential_parameter, max_contraction_val, tmp_max, eps_schwarz,&
                                            neris_tmp,Zeta1,ZetaInv,Eta,EtaInv,ZetapEtaInv,Rho,RhoInv,S1234,&
                                            P,Q,W,rpq2,AB,CD,&
                                            R1,R2)
                          cart_estimate = MAX(tmp_max,cart_estimate)
                        END DO !ld
                      END DO !lc
                    END DO !lb
                  END DO !la
                END DO !k
              END DO !lpgf
            END DO !kpgf
          END DO !jpgf
        END DO !ipgf
      END DO !j
    END DO !i
  END SUBROUTINE coulomb4_per

! *****************************************************************************
!> \brief Given a 2d index pair, this function returns a 1d index pair for
!>        a symmetric upper triangle NxN matrix
!>        The compiler should inline this function, therefore it appears in
!>        several modules
!> \param i,j 2d index
!> \param N matrix size
!> \par History
!>      03.2009 created [Manuel Guidon]
!> \author Manuel Guidon
! *****************************************************************************
  PURE FUNCTION get_1D_idx(i,j,N)
    INTEGER, INTENT(IN)                      :: i, j
    INTEGER(int_8), INTENT(IN)               :: N
    INTEGER(int_8)                           :: get_1D_idx

    INTEGER(int_8)                           :: min_ij

    min_ij = MIN(i,j)
    get_1D_idx = min_ij*N + MAX(i,j) - (min_ij-1)*min_ij/2 - N

  END FUNCTION get_1D_idx

! *****************************************************************************
!> \brief This routine prefetches density/fock matrix elements and stores them
!>        in cache friendly arrays. These buffers are then used to update the
!>        fock matrix
!> \param pa1, pb1, pc1, pd1 Offsets
!> \param ma_max, mb_max, mc_max, md_max: Size of matrix blocks
!> \param sgfa, sgfb sgfc, sgfd Offsets
!> \param fac, symm_fac mulitplication factors (spin, symmetry)
!> \param density upper triangular density matrix
!> \param ks upper triangular fock matrix
!> \param prim primitive integrals
!> \param pbd, pbc, pad, pac buffers that will contain P(b,d), P(b,c) ...
!> \param kbd, kbc, kad, kac buffers for KS(b,d), KS(b,c) ...
!> \param N_pmat 1D size of density/fock matrix
!> \par History
!>      03.2009 created [Manuel Guidon]
!> \author Manuel Guidon
! *****************************************************************************

  SUBROUTINE update_fock_matrix(pa1, pb1, pc1, pd1, ma_max, mb_max, mc_max, md_max, &
                                sgfa, sgfb, sgfc, sgfd, fac, symm_fac, density, ks, prim, &
                                pbd, pbc, pad, pac, kbd, kbc, kad, kac, N_pmat)

    INTEGER, INTENT(IN)                      :: pa1, pb1, pc1, pd1, ma_max, &
                                                mb_max, mc_max, md_max, sgfa, &
                                                sgfb, sgfc, sgfd
    REAL(dp), INTENT(IN)                     :: fac, symm_fac
    REAL(dp), DIMENSION(:)                   :: density, ks
    REAL(dp), DIMENSION(ma_max*mb_max*mc_max&
      *md_max), INTENT(IN)                   :: prim
    REAL(dp), DIMENSION(*)                   :: pbd, pbc, pad, pac, kbd, kbc, &
                                                kad, kac
    INTEGER(int_8), INTENT(IN)               :: N_pmat

    INTEGER                                  :: i, ma, mb, mc, md, p_index, &
                                                pa, pa2, pb, pb2, pc, pc2, &
                                                pd, pd2
    REAL(dp)                                 :: ks_bc, ks_bd, p_bc, p_bd, tmp

    i = 1
    DO md=1,md_max
      pd2 = md+sgfd-1
      pd = pd1 + pd2
      DO mb=1,mb_max
        pb2 = mb+sgfb-1
        pb = pb1 + pb2
        pbd(i) = density(get_1d_idx(pb,pd,N_pmat))
        i = i+1
      END DO
    END DO  
    i = 1
    DO mc=1,mc_max
      pc2 = mc+sgfc-1
      pc = pc1 + pc2
      DO mb=1,mb_max
        pb2 = mb+sgfb-1
        pb = pb1 + pb2
        pbc(i) = density(get_1d_idx(pb,pc,N_pmat))
        i = i+1
      END DO
    END DO  
    i = 1
    DO md=1,md_max
      pd2 = md+sgfd-1
      pd = pd1 + pd2
      DO ma=1,ma_max
        pa2 = ma+sgfa-1
        pa = pa1 + pa2
        pad(i) = density(get_1d_idx(pa,pd,N_pmat))
        i = i+1
      END DO
    END DO  
    i = 1
    DO mc=1,mc_max
      pc2 = mc+sgfc-1
      pc = pc1 + pc2
      DO ma=1,ma_max
        pa2 = ma+sgfa-1
        pa = pa1 + pa2
        pac(i) = density(get_1d_idx(pa,pc,N_pmat))
        i = i+1
      END DO
    END DO  

    kbd(1:mb_max*md_max) = 0.0_dp
    kbc(1:mb_max*mc_max) = 0.0_dp
    kad(1:ma_max*md_max) = 0.0_dp
    kac(1:ma_max*mc_max) = 0.0_dp

    p_index = 0
    DO md = 1,md_max
      DO mc = 1,mc_max
        DO mb = 1,mb_max
          ks_bd = 0.0_dp
          ks_bc = 0.0_dp
          p_bd = pbd((md-1)*mb_max+mb)
          p_bc = pbc((mc-1)*mb_max+mb)
          DO ma = 1,ma_max
            p_index=p_index+1
            tmp = fac*symm_fac*prim(p_index)
            ks_bc = ks_bc + tmp* pad((md-1)*ma_max+ma)
            ks_bd = ks_bd + tmp* pac((mc-1)*ma_max+ma)
            kad((md-1)*ma_max+ma) =  kad((md-1)*ma_max+ma)-tmp*p_bc
            kac((mc-1)*ma_max+ma) = kac((mc-1)*ma_max+ma)-tmp*p_bd
          END DO
          kbd((md-1)*mb_max+mb) = kbd((md-1)*mb_max+mb) - ks_bd
          kbc((mc-1)*mb_max+mb)=kbc((mc-1)*mb_max+mb) - ks_bc
        END DO
      END DO
    END DO


    i = 1
    DO md=1,md_max
      pd2 = md+sgfd-1
      pd = pd1 + pd2
      DO mb=1,mb_max
        pb2 = mb+sgfb-1
        pb = pb1 + pb2
        !$OMP ATOMIC
        ks(get_1d_idx(pb,pd,N_pmat)) = ks(get_1d_idx(pb,pd,N_pmat)) + kbd(i)
        i = i+1
      END DO
    END DO  
    i = 1
    DO mc=1,mc_max
      pc2 = mc+sgfc-1
      pc = pc1 + pc2
      DO mb=1,mb_max
        pb2 = mb+sgfb-1
        pb = pb1 + pb2
        !$OMP ATOMIC
        ks(get_1d_idx(pb,pc,N_pmat)) = ks(get_1d_idx(pb,pc,N_pmat)) + kbc(i) 
        i = i+1
      END DO
    END DO  
    i = 1
    DO md=1,md_max
      pd2 = md+sgfd-1
      pd = pd1 + pd2
      DO ma=1,ma_max
        pa2 = ma+sgfa-1
        pa = pa1 + pa2
        !$OMP ATOMIC
        ks(get_1d_idx(pa,pd,N_pmat)) = ks(get_1d_idx(pa,pd,N_pmat)) + kad(i)
        i = i+1
      END DO
    END DO  
    i = 1
    DO mc=1,mc_max
      pc2 = mc+sgfc-1
      pc = pc1 + pc2
      DO ma=1,ma_max
        pa2 = ma+sgfa-1
        pa = pa1 + pa2
        !$OMP ATOMIC
        ks(get_1d_idx(pa,pc,N_pmat)) = ks(get_1d_idx(pa,pc,N_pmat)) + kac(i)
        i = i+1
      END DO
    END DO  
  END SUBROUTINE update_fock_matrix
END MODULE hfx_energy_potential
